#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "mp_msg.h"
#include "help_mp.h"

#include "codec-cfg.h"
#include "stream/stream.h"
#include "libmpdemux/demuxer.h"
#include "libmpdemux/stheader.h"

#include "stream/stream.h"
#include "libmpdemux/muxer.h"

#include "img_format.h"
#include "mp_image.h"
#include "vf.h"
#include "libvo/fastmemcpy.h"


//===========================================================================//

struct vf_priv_s {
    muxer_stream_t* mux;
    mp_image_t*     timg;
};
#define mux_v (vf->priv->mux)
#define timg (vf->priv->timg)

static int set_format(struct vf_instance_s *vf, unsigned int fmt) {
    if (!force_fourcc)
        mux_v->bih->biCompression = fmt;
    
    mux_v->bih->biPlanes = 1;
    if (IMGFMT_IS_RGB(fmt)) {
	if (IMGFMT_RGB_DEPTH(fmt) < 8 && !(fmt&128))
	    mux_v->bih->biBitCount = IMGFMT_RGB_DEPTH(fmt);
	else
	    mux_v->bih->biBitCount = (IMGFMT_RGB_DEPTH(fmt)+7)&(~7);
	return 1;
    }
    if (IMGFMT_IS_BGR(fmt)) {
	if (IMGFMT_BGR_DEPTH(fmt) < 8 && !(fmt&128))
	    mux_v->bih->biBitCount = IMGFMT_BGR_DEPTH(fmt);
	else
	    mux_v->bih->biBitCount = (IMGFMT_BGR_DEPTH(fmt)+7)&(~7);
	return 1;
    }
    switch (fmt) {
    case IMGFMT_I420:
    case IMGFMT_IYUV:
    case IMGFMT_YV12:
    case IMGFMT_411P:
	mux_v->bih->biPlanes = 3;
	mux_v->bih->biBitCount = 12;
	break;
    case IMGFMT_444P:
	mux_v->bih->biPlanes = 3;
	mux_v->bih->biBitCount = 24;
	break;
    case IMGFMT_422P:
	mux_v->bih->biPlanes = 3;
	mux_v->bih->biBitCount = 16;
	break;
    case IMGFMT_IF09:
	mux_v->bih->biPlanes = 4;
    case IMGFMT_YVU9:
	mux_v->bih->biBitCount = 9;
	break;
    case IMGFMT_UYVY:
    case IMGFMT_YUY2:
	mux_v->bih->biBitCount = 16;
	break;
    case IMGFMT_Y8:
        mux_v->bih->biBitCount = 8;
        break;
    default:
	mp_msg(MSGT_MENCODER, MSGL_INFO, MSGTR_MPCODECS_OutputWithFourccNotSupported, fmt);
	mux_v->bih->biCompression = 0;
	return 0;
    }
    return 1;
}


static int config(struct vf_instance_s *vf,
        int width, int height, int d_width, int d_height,
	unsigned int flags, unsigned int outfmt)
{
    int ret;
    mux_v->bih->biWidth = width;
    mux_v->bih->biHeight = height;
    mux_v->aspect = (float)d_width/d_height;
    ret = set_format(vf, outfmt);
    if (!ret) return 0;

    mux_v->bih->biSizeImage = mux_v->bih->biWidth*mux_v->bih->biHeight*mux_v->bih->biBitCount/8;
    
    /* Allocate a buffer for a temproary image. Note that this code is ripped
       from vf_get_image. 
       
       Typically, vf_get_image is used to get an image from the _next_ filter
       in the chain as part of the DR mechanism. Since we're an encoder, 
       there is no next filter, thus there is no vf context to get an image
       from, thus we handle the allocation ourselves. 
       
       This code should probaby be put into its own function, since it's common
       to vf_get_image, this function, and vf_vo will need the same hack for
       draw_image.
    
       This may not be a permanant fix, pending input from other mplayer
       developers.
    */
    timg = new_mp_image(width,height);
    if(!timg) return 0;
    timg->type=MP_IMGTYPE_TEMP;
    mp_image_setfmt(timg,outfmt);
    timg->planes[0]=memalign(64, timg->bpp*timg->width*(timg->height+2)/8);
    if(timg->flags&MP_IMGFLAG_PLANAR){
	// YV12/I420/YVU9/IF09. feel free to add other planar formats here...
	//if(!timg->stride[0]) 
	timg->stride[0]=timg->width;
	//if(!timg->stride[1]) 
	if(timg->num_planes > 2){
	timg->stride[1]=timg->stride[2]=timg->chroma_width;
	if(timg->flags&MP_IMGFLAG_SWAPPED){
	    // I420/IYUV  (Y,U,V)
	    timg->planes[1]=timg->planes[0]+timg->width*timg->height;
	    timg->planes[2]=timg->planes[1]+timg->chroma_width*timg->chroma_height;
	} else {
	    // YV12,YVU9,IF09  (Y,V,U)
	    timg->planes[2]=timg->planes[0]+timg->width*timg->height;
	    timg->planes[1]=timg->planes[2]+timg->chroma_width*timg->chroma_height;
	}
	} else {
	    // NV12/NV21
	    timg->stride[1]=timg->chroma_width;
	    timg->planes[1]=timg->planes[0]+timg->width*timg->height;
	}
    } else {
	//if(!timg->stride[0]) 
	timg->stride[0]=timg->width*timg->bpp/8;
    }
    vf_mpi_clear(timg,0,0,timg->width,timg->height);
    timg->flags|=MP_IMGFLAG_ALLOCATED;
    
    return 1;
}

static int control(struct vf_instance_s *vf, int request, void *data) {
    return CONTROL_UNKNOWN;
}

static int query_format(struct vf_instance_s *vf, unsigned int fmt) {
    const int supported = VFCAP_CSP_SUPPORTED | VFCAP_CSP_SUPPORTED_BY_HW
        | VFCAP_ACCEPT_STRIDE;
        
    if (IMGFMT_IS_RGB(fmt) || IMGFMT_IS_BGR(fmt))
	return supported;
    switch (fmt) {
    case IMGFMT_I420:
    case IMGFMT_IYUV:
    case IMGFMT_YV12:
    case IMGFMT_411P:
    case IMGFMT_444P:
    case IMGFMT_422P:
    case IMGFMT_UYVY:
    case IMGFMT_YUY2:
    case IMGFMT_YVU9:
    case IMGFMT_IF09:
    case IMGFMT_Y8:
	return supported;
    }

    return 0;
}

static int put_image(struct vf_instance_s *vf, mp_image_t *mpi, double pts) {
    /* Pack the frame, if necessary. */
    if((timg->stride[0] != mpi->stride[0])
        || ((mpi->flags & MP_IMGFLAG_PLANAR) && (timg->height != mpi->height)))
    {

        memcpy_pic(timg->planes[0], mpi->planes[0], timg->stride[0],
                   timg->h, timg->stride[0], mpi->stride[0]);
        if (mpi->flags & MP_IMGFLAG_PLANAR) {
            memcpy_pic(timg->planes[1], mpi->planes[1], timg->stride[1],
                       timg->chroma_height, timg->stride[1], mpi->stride[1]);
            memcpy_pic(timg->planes[2], mpi->planes[2], timg->stride[2],
                       timg->chroma_height, timg->stride[2], mpi->stride[2]);
        }        
        mux_v->buffer = timg->planes[0];
    }
    else
        mux_v->buffer = mpi->planes[0];

    /* Write the packed frame to the muxer */
    muxer_write_chunk(mux_v, timg->width*timg->height*mux_v->bih->biBitCount/8, 0x10, pts, pts);
    return 1;
}

//===========================================================================//

static int vf_open(vf_instance_t *vf, char* args){
    vf->config = config;
    vf->default_caps = VFCAP_ACCEPT_STRIDE;
    vf->control = control;
    vf->query_format = query_format;
    vf->put_image = put_image;
    vf->priv = malloc(sizeof(struct vf_priv_s));
    memset(vf->priv, 0, sizeof(struct vf_priv_s));
    vf->priv->mux = (muxer_stream_t*)args;
    
    mux_v->bih = calloc(1, sizeof(BITMAPINFOHEADER));
    mux_v->bih->biSize = sizeof(BITMAPINFOHEADER);
    mux_v->bih->biWidth = 0;
    mux_v->bih->biHeight = 0;

    return 1;
}

vf_info_t ve_info_raw = {
    "raw encoder",
    "raw",
    "jwe21@cam.ac.uk",
    "Based on rawrgb",
    vf_open
};

//===========================================================================//
