/*******************************************************************************
*                                                                              *
* Copyright (c) 2005, On2 Technologies, Inc                                    *
* All rights reserved.                                                         *
*                                                                              *
* Redistribution and use in source and binary forms, with or without           *
* modification, are permitted provided that the following conditions are met:  *
*                                                                              *
*  - Redistributions of source code must retain the above copyright notice,    *
*    this list of conditions and the following disclaimer.                     *
*                                                                              *
*  - Redistributions in binary form must reproduce the above copyright notice, *
*    this list of conditions and the following disclaimer in the documentation *
*    and/or other materials provided with the distribution.                    *
*                                                                              *
*  - The names of the contributors may not be used to endorse or promote       *
*    products derived from this software without specific prior written        *
*    permission.                                                               *
*                                                                              *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  *
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    *
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   *
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE     *
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR          *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF         *
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS     *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN      *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)      *
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE   *
* POSSIBILITY OF SUCH DAMAGE.                                                  *
*                                                                              *
*******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ogg/ogg.h>
#include "oggpcm.h"

static int oggpcm_packetbuf_alloc(oggpcm_stream_t  *os, long size) {
    if(os->pkt_alloc_size)
        free(os->pkt_data_payload);
    os->pkt_data_payload = malloc(size);
    os->pkt_data_payload_sz = size;
    
    if(os->pkt_data_payload != NULL) {
        /* Update pointers based on new packet buffer */
        os->data_pkt.packet = os->pkt_data_payload; 
        os->pkt_alloc_size= size;                
    }
    
    return (os->pkt_data_payload == NULL);
}

static int oggpcm_fwrite_page(ogg_page *op, FILE *of) {
    int fail = 0;
    if((size_t)op->header_len != fwrite(op->header,1,op->header_len,of)) {
        fail = 1;
    }

    if(!fail && (size_t)op->body_len != fwrite(op->body,1,op->body_len,of)) {
        fail = 1;
    }
    return fail;
}

static int calc_bytes_per_frame(int channels, oggpcm_encoding_t fmt) {
    int val =0;
    
    switch(fmt) {
    /* Integer coding */
    case OGGPCM_FMT_S8:
    case OGGPCM_FMT_U8:
    case OGGPCM_FMT_ULAW:
    case OGGPCM_FMT_ALAW:                        
        val = 1 * channels;
        break;  
    case OGGPCM_FMT_S16_LE:
    case OGGPCM_FMT_S16_BE:
        val = 2 * channels;
        break;  
    case OGGPCM_FMT_S24_LE:
    case OGGPCM_FMT_S24_BE:
        val = 3 * channels;
        break;  
    case OGGPCM_FMT_FLT32_LE:          
    case OGGPCM_FMT_FLT32_BE:        
    case OGGPCM_FMT_S32_LE:  
    case OGGPCM_FMT_S32_BE:
        val = 4 * channels;
        break;  
    case OGGPCM_FMT_FLT64_LE:      
    case OGGPCM_FMT_FLT64_BE:
        val = 8 * channels;
        break;  
    }
    return val;            
}

int oggpcm_header_pageout (oggpcm_stream_t *os, FILE *of) {
    oggpack_buffer hdr_buf;
    ogg_packet     hdr_pkt;
    ogg_page       op;
    int            result;
    
    /* Pack the header data into the buffer. */
    oggpackB_writeinit(&hdr_buf);
    oggpackB_write(&hdr_buf, 'P', 8);
    oggpackB_write(&hdr_buf, 'C', 8);
    oggpackB_write(&hdr_buf, 'M', 8);
    oggpackB_write(&hdr_buf, ' ', 8);
    oggpackB_write(&hdr_buf, ' ', 8);
    oggpackB_write(&hdr_buf, ' ', 8);
    oggpackB_write(&hdr_buf, ' ', 8);
    oggpackB_write(&hdr_buf, ' ', 8);
    oggpackB_write(&hdr_buf, 0x00, 16); /* Version Major */
    oggpackB_write(&hdr_buf, 0x00, 16); /* Version Minor */
    oggpackB_write(&hdr_buf, os->fmt, 32);
    oggpackB_write(&hdr_buf, os->frames_per_sec, 32);
    oggpackB_write(&hdr_buf, os->sig_bits, 8);    
    oggpackB_write(&hdr_buf, os->channels, 8);
    oggpackB_write(&hdr_buf, os->max_frames_per_pkt, 16);
    oggpackB_write(&hdr_buf, 0, 32); /* Number of extra headers */
    assert(oggpack_bits(&hdr_buf) == 224);
    
    /* Pack the buffer into a packet */
    hdr_pkt.b_o_s = 1;
    hdr_pkt.e_o_s = 0;
    hdr_pkt.granulepos = os->granulepos;
    hdr_pkt.packetno = os->packetno++;
    hdr_pkt.packet = oggpack_get_buffer(&hdr_buf);
    hdr_pkt.bytes = oggpack_bytes(&hdr_buf);
    
    /* Insert the packet into the stream */
    result = ogg_stream_packetin(&os->ogg_ss, &hdr_pkt);
    assert(result == 0);
    
    /* The header packet should be the only packet on the page, so
    ** flush it rather than pageout.
    */
    result = ogg_stream_flush(&os->ogg_ss, &op);
    assert(result != 0);
    result = !result;
    assert(result == 0);
    
    /* Write the page to the file. */
    if(result == 0)
        result = oggpcm_fwrite_page(&op, of);
    
    /* Clean up. */
    oggpackB_writeclear(&hdr_buf);  
    return result;
}

int oggpcm_footer_pageout (oggpcm_stream_t *os, FILE *of) {
    ogg_packet     hdr_pkt;
    ogg_page       op;
    int            result;
    int again;

    /* flush any remaining data packets before adding e_o_s */
    result = oggpcm_stream_flush(os,of); assert(result>=0);
        
    /* Assemble packet */
    memset(&hdr_pkt, 0, sizeof(ogg_packet));    
    hdr_pkt.b_o_s = 0;
    hdr_pkt.e_o_s = 1;
    hdr_pkt.granulepos = os->granulepos + 1;
    hdr_pkt.packetno = os->packetno++;
    hdr_pkt.packet = (unsigned char*)&hdr_pkt;
    hdr_pkt.bytes = 0;
    
    /* Insert the packet into the stream */
    result = ogg_stream_packetin(&os->ogg_ss, &hdr_pkt);
    assert(result == 0);
    
    /* The footer packet should be the only packet on the page, so
    ** flush it rather than pageout.
    */
    result = -1;
    do {
        again = ogg_stream_flush(&os->ogg_ss, &op);
        /* Write the page to the file. */
        if(again)
            result = oggpcm_fwrite_page(&op, of);
    } while(!result && again);
    
    return result;
}

int oggpcm_stream_packetin (oggpcm_stream_t  *os, const void *buf, long buf_size, double time) {
    unsigned char *bufp   = (unsigned char *)buf;
    long           frames_togo;
    int            result = 0;
    long           size;
    long           frames_this_pkt;
        
    frames_togo = buf_size / os->bytes_per_frame;
    os->granulepos = time * os->frames_per_sec;
    
    while(frames_togo && !result) {
        /*
        ** Calculate the packet size to use for this packet. Should be
        ** the largest number of full frames that can fit in 1k.
        */
        frames_this_pkt = (os->max_frames_per_pkt < frames_togo)?
            os->max_frames_per_pkt : frames_togo;
        size = frames_this_pkt * os->bytes_per_frame;
        
        result = (buf == NULL);
        if(!result && size > os->pkt_alloc_size) 
            result = oggpcm_packetbuf_alloc(os, size);

        if(!result) {
            memcpy(os->pkt_data_payload, bufp, size);
            os->data_pkt.granulepos = os->granulepos;
            os->data_pkt.packetno = os->packetno++;
            os->data_pkt.bytes = size;
            ogg_stream_packetin(&os->ogg_ss, &os->data_pkt);

            os->granulepos = os->data_pkt.granulepos + frames_this_pkt;
            bufp += size;
        }
        
        frames_togo -= frames_this_pkt;
    }
    return result;
}

int oggpcm_stream_pageout (oggpcm_stream_t  *os, FILE *of) {
    ogg_page op;
    int fail = 0;
    int pages = 0;
    
    while(ogg_stream_pageout(&os->ogg_ss, &op)) {
        pages++;
        fail = oggpcm_fwrite_page(&op, of);
        if(fail) {
            pages = -pages;
            break;
        }
    }
    
    return pages;
}
   
int oggpcm_stream_flush (oggpcm_stream_t  *os, FILE *of) {
    ogg_page op;
    int fail = 0;
    int pages = 0;
    
    while(ogg_stream_flush(&os->ogg_ss, &op)) {
        pages++;
        fail = oggpcm_fwrite_page(&op, of);
        if(fail) {
            pages = -pages;
            break;
        }
    }
    
    return pages;
}

int oggpcm_stream_parse_pkt (oggpcm_stream_t *os, ogg_packet *op) {
    oggpack_buffer buf;
    int result = -1;
    
    /* Stream has not been initialized yet. Try to parse a header packet */
    if(op->bytes >= 28) {
        oggpackB_readinit(&buf, op->packet, op->bytes);
        if(oggpackB_read(&buf, 8) == 'P' &&
            oggpackB_read(&buf, 8) == 'C' &&
            oggpackB_read(&buf, 8) == 'M' &&
            oggpackB_read(&buf, 8) == ' ' && 
            oggpackB_read(&buf, 8) == ' ' && 
            oggpackB_read(&buf, 8) == ' ' && 
            oggpackB_read(&buf, 8) == ' ' && 
            oggpackB_read(&buf, 8) == ' ' &&
            oggpackB_read(&buf, 16) == 0x0000) {

            /* Looks like a valid OggPCM v0.x header */
            memset(os, 0, sizeof(oggpcm_stream_t));
            oggpackB_adv(&buf, 16);
            os->fmt = oggpackB_read(&buf, 32);
            os->frames_per_sec = oggpackB_read(&buf, 32);
            os->sig_bits = oggpackB_read(&buf, 8);
            os->channels = oggpackB_read(&buf, 8);
            os->max_frames_per_pkt = oggpackB_read(&buf, 16);
            os->num_extra_headers = oggpackB_read(&buf, 32);
            result = 0;
            os->bytes_per_frame = calc_bytes_per_frame(os->channels, os->fmt);
        }
        
        /* Skip any extra header packets */
        if(!op->b_o_s && os->num_extra_headers_read < os->num_extra_headers) {
            os->num_extra_headers_read++;
            result = 0;
        }
        
        oggpackB_readinit(&buf, op->packet, op->bytes);
        if(result == -1
            && os->bytes_per_frame > 0 
            && op->bytes % os->bytes_per_frame == 0) {
            
            /* Looks like a valid data packet */
            memcpy(&os->data_pkt, op, sizeof(ogg_packet));
            os->pkt_data_payload = op->packet;
            os->pkt_data_payload_sz = op->bytes;
            os->packetno = op->packetno;
            if(op->granulepos == -1) {
                /* Need to fix up granulepos. */
                op->granulepos = os->granulepos;
                os->granulepos += (op->bytes-4)/os->bytes_per_frame;
            }
            else
                os->granulepos = op->granulepos;
    
            
            result = 1;        
        }
    }    
    return result;
}

int oggpcm_stream_init (oggpcm_stream_t  *os,                 
                        int               channels,           
                        int               sig_bits,    
                        int               frames_per_sec,    
                        oggpcm_encoding_t fmt,                
                        int               serialno) {
    (void)sig_bits;
    memset(os, 0, sizeof(oggpcm_stream_t));
    os->channels = channels;
    os->frames_per_sec = frames_per_sec;
    os->fmt = fmt;
    os->bytes_per_frame = calc_bytes_per_frame(channels, fmt);
    os->max_frames_per_pkt = 1020 / os->bytes_per_frame;
    
    os->data_pkt.b_o_s = 0;
    os->data_pkt.e_o_s = 0;
    os->packetno = 0;
    os->granulepos = 0;
    os->pkt_alloc_size = 0;
    return ogg_stream_init(&os->ogg_ss, serialno);
}

void oggpcm_stream_destroy(oggpcm_stream_t *os)
{
    ogg_stream_clear(&os->ogg_ss);
    free(os->pkt_data_payload);
    os->pkt_data_payload = NULL;
}
