/*******************************************************************************
*                                                                              *
* Copyright (c) 2005, On2 Technologies, Inc                                    *
* All rights reserved.                                                         *
*                                                                              *
* Redistribution and use in source and binary forms, with or without           *
* modification, are permitted provided that the following conditions are met:  *
*                                                                              *
*  - Redistributions of source code must retain the above copyright notice,    *
*    this list of conditions and the following disclaimer.                     *
*                                                                              *
*  - Redistributions in binary form must reproduce the above copyright notice, *
*    this list of conditions and the following disclaimer in the documentation *
*    and/or other materials provided with the distribution.                    *
*                                                                              *
*  - The names of the contributors may not be used to endorse or promote       *
*    products derived from this software without specific prior written        *
*    permission.                                                               *
*                                                                              *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  *
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    *
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   *
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE     *
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR          *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF         *
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS     *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN      *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)      *
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE   *
* POSSIBILITY OF SUCH DAMAGE.                                                  *
*                                                                              *
*******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <ogg/ogg.h>
#include "ogguvs.h"

static void rat(double f, int *num, int *den) {
    int intg, n0=0, d0=0, n1=1, n2=0, d1=0, d2=1;
    double quot,frac;

    quot = f;
    for(;;) {
        intg=(int)(quot);
        frac=quot-(double)intg;
        if(n1*intg+n2 > 65535) break;
        if(d1*intg+d2 > 65535) break;
        n0=n1*intg+n2;
        d0=d1*intg+d2;
        /* Rotate */
        n2=n1; n1=n0;
        d2=d1; d1=d0;
        if(frac < 0.00001)
            break;
        else
            quot = 1.0/frac;
    };

    *num = n0;
    *den = d0;

}

static int ogguvs_packetbuf_alloc(ogguvs_stream_t  *os, long size) {

    if(os->pkt_alloc_size)
        free(os->pkt_data);
    os->pkt_data = malloc(size + 12);

    if(os->pkt_data != NULL) {
        /* Update pointers based on new packet buffer */
        os->data_pkt.packet = os->pkt_data;
        os->pkt_data_payload = os->pkt_data + 12;
        os->pkt_data_payload_sz = size;
        os->pkt_alloc_size= size;

    }

    return (os->pkt_data == NULL);
}

static int ogguvs_fwrite_page(ogg_page *op, FILE *of) {
    int fail = 0;
    if((size_t)op->header_len != fwrite(op->header,1,op->header_len,of)) {
        fail = 1;
    }

    if(!fail && (size_t)op->body_len != fwrite(op->body,1,op->body_len,of)) {
        fail = 1;
    }
    return fail;
}

int ogguvs_header_pageout (ogguvs_stream_t *os, FILE *of) {
    oggpack_buffer hdr_buf;
    ogg_packet     hdr_pkt;
    ogg_page       op;
    int            result;
    unsigned long  lng;

    /* Pack the header data into the buffer. */
    oggpackB_writeinit(&hdr_buf);
    oggpackB_write(&hdr_buf, 'U', 8);
    oggpackB_write(&hdr_buf, 'V', 8);
    oggpackB_write(&hdr_buf, 'S', 8);
    oggpackB_write(&hdr_buf, 0, 8);
    oggpackB_write(&hdr_buf, 0, 32);

    oggpackB_write(&hdr_buf, 0x01, 16); /* Version Major */
    oggpackB_write(&hdr_buf, 0x00, 16); /* Version Minor */
    oggpackB_write(&hdr_buf, os->disp_width, 16);
    oggpackB_write(&hdr_buf, os->disp_height, 16);
    oggpackB_write(&hdr_buf, os->pix_aspect_num, 16);
    oggpackB_write(&hdr_buf, os->pix_aspect_den, 16);
    oggpackB_write(&hdr_buf, os->fld_rate_num, 16);
    oggpackB_write(&hdr_buf, os->fld_rate_den, 16);
    oggpackB_write(&hdr_buf, os->timebase, 32);
    oggpackB_write(&hdr_buf, os->field_size, 32);
    oggpackB_write(&hdr_buf, 0, 32); /* Number of extra headers */
    lng = (unsigned long)os->colorspace;
    oggpackB_write(&hdr_buf, lng, 32); /* Colorspace */
    oggpackB_write(&hdr_buf, os->interlaced, 32);
    oggpackB_write(&hdr_buf, os->layout_id, 32);

    assert(oggpack_bits(&hdr_buf) == 12*32);

    /* Pack the buffer into a packet */
    hdr_pkt.b_o_s = 1;
    hdr_pkt.e_o_s = 0;
    hdr_pkt.granulepos = os->granulepos++;
    hdr_pkt.packetno = os->packetno++;
    hdr_pkt.packet = oggpack_get_buffer(&hdr_buf);
    hdr_pkt.bytes = oggpack_bytes(&hdr_buf);

    /* Insert the packet into the stream */
    result = ogg_stream_packetin(&os->ogg_ss, &hdr_pkt);
    assert(result == 0);

    /* The header packet should be the only packet on the page, so
    ** flush it rather than pageout.
    */
    result = ogg_stream_flush(&os->ogg_ss, &op);
    assert(result != 0);
    result = !result;
    assert(result == 0);

    /* Write the page to the file. */
    if(result == 0)
        result = ogguvs_fwrite_page(&op, of);

    /* Clean up. */
    oggpackB_writeclear(&hdr_buf);
    return result;
}

int ogguvs_footer_pageout (ogguvs_stream_t *os, FILE *of) {
    ogg_packet     hdr_pkt;
    ogg_page       op;
    int            result;
    int again;

    /* as the comment below suggests, if we really want this to be the only
       packet on the page, we should flush prior to adding it */
    result = ogguvs_stream_flush(os,of); assert(result>=0);

    /* Assemble packet */
    memset(&hdr_pkt, 0, sizeof(ogg_packet));
    hdr_pkt.b_o_s = 0;
    hdr_pkt.e_o_s = 1;
    hdr_pkt.granulepos = os->granulepos + 1;
    hdr_pkt.packetno = os->packetno++;
    hdr_pkt.packet = (unsigned char*)&hdr_pkt;
    hdr_pkt.bytes = 0;

    /* Insert the packet into the stream */
    result = ogg_stream_packetin(&os->ogg_ss, &hdr_pkt);
    assert(result == 0);

    /* The footer packet should be the only packet on the page, so
    ** flush it rather than pageout.
    */
    result = -1;
    do {
        again = ogg_stream_flush(&os->ogg_ss, &op);
        /* Write the page to the file. */
        if(again)
            result = ogguvs_fwrite_page(&op, of);
    } while(!result && again);

    return result;
}
int ogguvs_stream_packetin (ogguvs_stream_t  *os,
                            ogguvs_fieldid_t  fld,
                            const void       *buf,
                            long              size,
                            double            time) {
    int            result = 0;
    oggpack_buffer hdr_buf;
    ogg_int64_t    field_time;

    result = ((buf == NULL)||(size != os->field_size));

    if(!result && size > os->pkt_alloc_size)
        result = ogguvs_packetbuf_alloc(os, size);

    if(!size && !time) {
        /* Material is fixed fps (indicated by time==0), but passed a 0 sized
         * frame. In the AVI world, this means duplicate last frame. We don't
         * do that, but we do need to increment the packet number so that the
         * timestamps will be calculated correctly.
         */
        os->packetno++;
    }

    if(!result) {
        /* Calculate field time in the specified timebase */
        if(os->timebase && (time > 0))
            field_time = (ogg_int64_t)(os->timebase * time);
        else if (os->timebase && os->fld_rate_num)
            /* packetno-1 is the current frame number, since there is 1
            ** header packet.
            **/
            field_time = (os->packetno-1) * os->timebase
                * os->fld_rate_den / os->fld_rate_num;
        else
            field_time = os->granulepos + 1;

        /* Pack a data packet header into the new packet buffer */
        oggpackB_writeinit(&hdr_buf);
        oggpackB_write(&hdr_buf, 'F', 8);
        oggpackB_write(&hdr_buf, 'L', 8);
        oggpackB_write(&hdr_buf, 'D', 8);
        oggpackB_write(&hdr_buf, '0'+fld, 8);
        oggpackB_write(&hdr_buf, (unsigned long)(field_time >> 32), 32);
        oggpackB_write(&hdr_buf, (unsigned long)(field_time & 0xFFFFFFFF), 32);
        assert(oggpack_bits(&hdr_buf) == 96);
        memcpy(os->pkt_data, oggpack_get_buffer(&hdr_buf), 12);
        oggpackB_writeclear(&hdr_buf);

        memcpy(os->pkt_data_payload, buf, size);
        os->data_pkt.packetno = ++os->packetno;
        os->data_pkt.bytes = 12 + size;
        os->data_pkt.granulepos = os->granulepos = field_time;
        ogg_stream_packetin(&os->ogg_ss, &os->data_pkt);
    }

    return result;
}

int ogguvs_stream_pageout (ogguvs_stream_t  *os, FILE *of) {
    ogg_page op;
    int fail = 0;
    int pages = 0;

    while(ogg_stream_pageout(&os->ogg_ss, &op)) {
        pages++;
        fail = ogguvs_fwrite_page(&op, of);
        if(fail) {
            pages = -pages;
            break;
        }
    }

    return pages;
}

int ogguvs_stream_flush (ogguvs_stream_t  *os, FILE *of) {
    ogg_page op;
    int fail = 0;
    int pages = 0;

    while(ogg_stream_flush(&os->ogg_ss, &op)) {
        pages++;
        fail = ogguvs_fwrite_page(&op, of);
        if(fail) {
            pages = -pages;
            break;
        }
    }

    return pages;
}

int ogguvs_stream_init (ogguvs_stream_t  *os,
                        int               disp_width,
                        int               disp_height,
                        double            pixel_aspect,
                        double            field_rate,
                        long              timebase,
                        long              field_size,
                        ogguvs_colorsp_t  colorspace,
                        ogguvs_format_t   layout_id,
                        long              serialno) {

    /* Validate parameters */
    if(field_rate < 0.0) return -1;

    os->disp_width = disp_width;
    os->disp_height = disp_height;

    if(pixel_aspect)
        rat(pixel_aspect, &os->pix_aspect_num, &os->pix_aspect_den);
    else
        os->pix_aspect_num = os->pix_aspect_den = 0;

    if(field_rate)
        rat(field_rate, &os->fld_rate_num, &os->fld_rate_den);
    else
        os->fld_rate_num = os->fld_rate_den = 0;

    os->field_size = field_size;
    os->colorspace = colorspace;
    os->layout_id = layout_id;
    os->interlaced = 0;
    os->timebase = timebase;
    os->data_pkt.b_o_s = 0;
    os->data_pkt.e_o_s = 0;
    os->packetno = 0;
    os->granulepos = 0;
    os->pkt_alloc_size = 0;
    return ogg_stream_init(&os->ogg_ss, serialno);
}

void ogguvs_stream_destroy(ogguvs_stream_t  *os)
{
    ogg_stream_clear(&os->ogg_ss);
    free(os->pkt_data);
    os->pkt_data = NULL;
}

int ogguvs_stream_parse_pkt (ogguvs_stream_t *os, ogg_packet *op) {
    oggpack_buffer buf;
    ogg_int64_t    field_time;
    int result = -1;

    /* Stream has not been initialized yet. Try to parse a header packet */
    if(op->bytes >= 48) {
        oggpackB_readinit(&buf, op->packet, op->bytes);
        if(oggpackB_read(&buf, 8) == 'U' &&
            oggpackB_read(&buf, 8) == 'V' &&
            oggpackB_read(&buf, 8) == 'S' &&
            oggpackB_read(&buf, 8) == 0x00 &&
            oggpackB_read(&buf, 32) == 0x00 &&
            oggpackB_read(&buf, 16) == 0x01) {

            /* Looks like a valid OggUVS v1.x header */
            memset(os, 0, sizeof(ogguvs_stream_t));
            oggpackB_adv(&buf, 16); /* Minor version */
            os->disp_width = oggpackB_read(&buf, 16);
            os->disp_height = oggpackB_read(&buf, 16);
            os->pix_aspect_num = oggpackB_read(&buf, 16);
            os->pix_aspect_den = oggpackB_read(&buf, 16);
            os->fld_rate_num = oggpackB_read(&buf, 16);
            os->fld_rate_den = oggpackB_read(&buf, 16);
            os->timebase = oggpackB_read(&buf, 32);
            os->field_size = oggpackB_read(&buf, 32);
            oggpackB_adv(&buf, 32); /* Number of extra headers */
            os->colorspace = oggpackB_read(&buf, 32);
            oggpackB_adv(&buf, 31);
            os->interlaced = oggpackB_read(&buf, 1);
            os->layout_id = oggpackB_read(&buf, 32);

            result = 0;
        }

        oggpackB_readinit(&buf, op->packet, op->bytes);
        if(result == -1 &&
            oggpackB_read(&buf, 8) == 'F' &&
            oggpackB_read(&buf, 8) == 'L' &&
            oggpackB_read(&buf, 8) == 'D') {

            /* Looks like a valid data packet */
            oggpackB_adv(&buf, 8);
            field_time = ((ogg_int64_t)(oggpackB_read(&buf, 32)) << 32);
            field_time |= (ogg_int64_t)(oggpackB_read(&buf, 32));

            memcpy(&os->data_pkt, op, sizeof(ogg_packet));
            os->pkt_data = op->packet;
            os->pkt_data_payload = os->pkt_data + 12;
            os->pkt_data_payload_sz = op->bytes - 12;
            os->packetno = op->packetno;
            if(op->granulepos == -1) {
                /* Need to fix up granulepos. */
                op->granulepos = ++os->granulepos;
            }
            else
                os->granulepos = op->granulepos;

            result = 1;
        }
    }
    return result;
}

